// Clavier+
// Keyboard shortcuts manager
//
// Copyright (C) 2000-2008 Guillaume Ryder
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#include "StdAfx.h"
#include "Global.h"

#include <intshcut.h>
#include <tlhelp32.h>
#include <psapi.h>

#include <dlgs.h>
#undef psh1
#undef psh2

const int kWindowClassBufferLength = 200;

HANDLE e_hHeap;
HINSTANCE e_hInst;
HWND e_hwndInvisible;
HWND e_hdlgModal;
bool e_bIconVisible = true;

#define myGetProcAddress(hDLL, functionName) \
	((PFN_##functionName)GetProcAddress(hDLL, (#functionName ANSI_UNICODE("A", "W"))))


static bool pathIsSlow(LPCTSTR path);

static BOOL CALLBACK prcEnumFindWindowByName(HWND hwnd, LPARAM lParam);


int messageBox(HWND hwnd, UINT idString, UINT uType, LPCTSTR pszArg) {
	TCHAR pszFormat[256], pszText[1024];
	i18n::loadStringAuto(idString, pszFormat);
	wsprintf(pszText, pszFormat, pszArg);
	return MessageBox(hwnd, pszText, pszApp, uType);
}


void centerParent(HWND hwnd) {
	RECT rcParent, rcChild;
	
	const HWND hwnd_parent = GetParent(hwnd);
	MONITORINFO mi;
	mi.cbSize = sizeof(mi);
	if (hwnd_parent) {
		const HMONITOR hMonitor = MonitorFromWindow(hwnd_parent, MONITOR_DEFAULTTONULL);
		VERIFV(hMonitor);
		GetMonitorInfo(hMonitor, &mi);
		GetWindowRect(hwnd_parent, &rcParent);
		IntersectRect(&rcParent, &rcParent, &mi.rcWork);
	} else {
		const HMONITOR hMonitor = MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST);
		VERIFV(hMonitor);
		GetMonitorInfo(hMonitor, &mi);
		rcParent = mi.rcWork;
	}
	
	GetWindowRect(hwnd, &rcChild);
	SetWindowPos(hwnd, NULL,
		(rcParent.left + rcParent.right + rcChild.left - rcChild.right) / 2,
		(rcParent.top + rcParent.bottom + rcChild.top - rcChild.bottom) / 2,
		0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}


void getDlgItemText(HWND hdlg, UINT id, String& rs) {
	const HWND hwnd = GetDlgItem(hdlg, id);
	const int buf = GetWindowTextLength(hwnd) + 1;
	GetWindowText(hwnd, rs.getBuffer(buf), buf);
}


static WNDPROC s_prcLabel;
static LRESULT CALLBACK prcWebLink(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

void initializeWebLink(HWND hdlg, UINT control_id, LPCTSTR link) {
	const HWND hwnd = GetDlgItem(hdlg, control_id);
	setWindowLongPtr(hwnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(link));
	s_prcLabel = subclassWindow(hwnd, prcWebLink);
}

// Window procedure for dialog box controls displaying an URL.
// The link itself is stored in GWL_USERDATA.
LRESULT CALLBACK prcWebLink(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
		
		case WM_SETCURSOR:
			SetCursor(LoadCursor(NULL, IDC_HAND));
			return TRUE;
		
		case WM_NCHITTEST:
			return HTCLIENT;
		
		case WM_LBUTTONDOWN:
			ShellExecute(hwnd, NULL, reinterpret_cast<LPCTSTR>(getWindowLongPtr(hwnd, GWLP_USERDATA)),
				NULL, NULL, SW_SHOWDEFAULT);
			return 0;
	}
	
	return CallWindowProc(s_prcLabel, hwnd, uMsg, wParam, lParam);
}


// Wrapper for CreateThread().
void startThread(LPTHREAD_START_ROUTINE pfn, void* params) {
	DWORD idThread;
	CloseHandle(CreateThread(NULL, 0, pfn, params, 0, &idThread));
}


void writeFile(HANDLE hf, LPCTSTR strbuf) {
	DWORD len;
	WriteFile(hf, strbuf, lstrlen(strbuf) * sizeof(TCHAR), &len, NULL);
}


bool getWindowProcessName(HWND hwnd, LPTSTR process_name) {
	DWORD process_id;
	GetWindowThreadProcessId(hwnd, &process_id);
	
	const HANDLE process_handle = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, process_id);
	VERIF(process_handle);
	const bool ok = (GetProcessImageFileName(process_handle, process_name, MAX_PATH) > 0);
	CloseHandle(process_handle);
	
	if (ok) {
		PathStripPath(process_name);
		CharLower(process_name);
	}
	
	return ok;
}


void sleepBackground(DWORD dwDurationMS) {
	const HANDLE hProcess = GetCurrentProcess();
	const DWORD dwOldPriority = GetPriorityClass(hProcess);
	SetPriorityClass(hProcess, IDLE_PRIORITY_CLASS);
	Sleep(dwDurationMS);
	SetPriorityClass(hProcess, dwOldPriority);
}


// Indicates if a file is located in a slow drive (network, removable, etc.).
bool pathIsSlow(LPCTSTR path) {
	const int drive_index = PathGetDriveNumber(path);
	if (drive_index >= 0) {
		// The path has a drive
		TCHAR root[4];
		PathBuildRoot(root, drive_index);
		switch (GetDriveType(root)) {
			case DRIVE_UNKNOWN:
			case DRIVE_REMOVABLE:
			case DRIVE_REMOTE:
			case DRIVE_CDROM:
				return true;
		}
	} else if (path[0] == _T('\\') && path[1] == _T('\\')) {
		// Network path
		return true;
	}
	return false;
}


bool getFileInfo(LPCTSTR path, DWORD file_attributes, SHFILEINFO& shfi, UINT flags) {
	TCHAR pszPathTemp[MAX_PATH];
	
	if (pathIsSlow(path)) {
		// If UNC server or share, remove any trailing backslash
		// If normal file, strip the path: extension-based icon
		
		flags |= SHGFI_USEFILEATTRIBUTES;
		lstrcpy(pszPathTemp, path);
		PathRemoveBackslash(pszPathTemp);
		const LPCTSTR pszPathName = PathFindFileName(path);
		if (PathIsUNCServer(pszPathTemp) || PathIsUNCServerShare(pszPathTemp)) {
			path = pszPathTemp;
		} else if (pszPathName) {
			lstrcpy(pszPathTemp, pszPathName);
			path = pszPathTemp;
		}
	}
	
	for (;;) {
		const bool bOK = toBool(SHGetFileInfo(path, file_attributes,
				&shfi, sizeof(shfi), flags));
		if (bOK || (flags & SHGFI_USEFILEATTRIBUTES)) {
			return bOK;
		}
		flags |= SHGFI_USEFILEATTRIBUTES;
	}
}


void clipboardToEnvironment() {
	bool ok = false;
	
	const UINT clipboard_format = ANSI_UNICODE(CF_TEXT, CF_UNICODETEXT);
	
	if (IsClipboardFormatAvailable(clipboard_format) && OpenClipboard(NULL)) {
		const HGLOBAL clipboard_mem = static_cast<HGLOBAL>(GetClipboardData(clipboard_format));
		if (clipboard_mem) {
			const LPTSTR clipboard_text = static_cast<LPTSTR>(GlobalLock(clipboard_mem));
			if (clipboard_text) {
				if (lstrlen(clipboard_text) < bufClipboardString) {
					ok = true;
					SetEnvironmentVariable(clipboard_env_variable, clipboard_text);
				}
				GlobalUnlock(clipboard_mem);
			}
		}
		CloseClipboard();
	}
	
	if (!ok) {
		SetEnvironmentVariable(clipboard_env_variable, _T(""));
	}
}


HWND findVisibleChildWindow(HWND hwnd_parent, LPCTSTR wnd_class, bool allow_same_prefix) {
	HWND hwndChild = NULL;
	while (toBool(hwndChild = FindWindowEx(hwnd_parent, hwndChild, NULL, NULL))) {
		if ((GetWindowStyle(hwndChild) & WS_VISIBLE) &&
				checkWindowClass(hwndChild, wnd_class, allow_same_prefix)) {
			return hwndChild;
		}
	}
	return NULL;
}

bool checkWindowClass(HWND hwnd, LPCTSTR wnd_class, bool allow_same_prefix) {
	TCHAR actual_wnd_class[kWindowClassBufferLength];
	VERIF(GetClassName(hwnd, actual_wnd_class, arrayLength(actual_wnd_class)));
	if (allow_same_prefix) {
		actual_wnd_class[lstrlen(wnd_class)] = _T('\0');
	}
	return !lstrcmp(wnd_class, actual_wnd_class);
}

struct FIND_WINDOW_BY_NAME {
	LPCTSTR title_regexp;
	HWND hwnd_found;
};

HWND findWindowByName(LPCTSTR title_regexp) {
	const LPTSTR pszWindowSpecLowercase = new TCHAR[lstrlen(title_regexp) + 1];
	lstrcpy(pszWindowSpecLowercase, title_regexp);
	CharLower(pszWindowSpecLowercase);
	
	FIND_WINDOW_BY_NAME fwbn;
	fwbn.title_regexp = pszWindowSpecLowercase;
	fwbn.hwnd_found = NULL;
	EnumWindows(prcEnumFindWindowByName, (LPARAM)&fwbn);
	delete [] pszWindowSpecLowercase;
	return fwbn.hwnd_found;
}

BOOL CALLBACK prcEnumFindWindowByName(HWND hwnd, LPARAM lParam) {
	FIND_WINDOW_BY_NAME &fwbn = *(FIND_WINDOW_BY_NAME*)lParam;
	TCHAR title[1024];
	if (GetWindowText(hwnd, title, arrayLength(title))) {
		if (matchWildcards(fwbn.title_regexp, title)) {
			fwbn.hwnd_found = hwnd;
			return FALSE;
		}
	}
	return TRUE;
}


bool matchWildcards(LPCTSTR pattern, LPCTSTR subject, LPCTSTR pattern_end) {
	for (;;) {
		const TCHAR pattern_chr = (pattern == pattern_end) ? _T('\0') : *pattern++;
		switch (pattern_chr) {
			
			case _T('\0'):
				return !*subject;
			
			case _T('*'):
				if (!*pattern || pattern == pattern_end) {
					return true;  // Optimization
				}
				do {
					if (matchWildcards(pattern, subject, pattern_end)) {
						return true;
					}
				} while (*subject++);
				return false;
			
			case _T('?'):
				VERIF(*subject++);
				break;
			
			default:
				VERIF(*subject++ == pattern_chr);
				break;
		}
	}
}


//------------------------------------------------------------------------
// SHBrowseForFolder wrapper:
// - use given title
// - set initial directory
// - return directory path instead of LPITEMIDLIST
//------------------------------------------------------------------------

static int CALLBACK prcBrowseForFolderCallback(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData);

bool browseForFolder(HWND hwnd_parent, LPCTSTR title, LPTSTR directory) {
	BROWSEINFO bi;
	ZeroMemory(&bi, sizeof(bi));
	bi.hwndOwner = hwnd_parent;
	bi.lpszTitle = title;
	bi.ulFlags = BIF_RETURNONLYFSDIRS;
	bi.lpfn = prcBrowseForFolderCallback;
	bi.lParam = (LPARAM)directory;
	LPITEMIDLIST pidl = SHBrowseForFolder(&bi);
	if (!pidl || !SHGetPathFromIDList(pidl, directory)) {
		return false;
	}
	CoTaskMemFree(pidl);
	return true;
}

int CALLBACK prcBrowseForFolderCallback(HWND hwnd, UINT uMsg, LPARAM, LPARAM lpData) {
	if (uMsg == BFFM_INITIALIZED) {
		SendMessage(hwnd, BFFM_SETSELECTION, TRUE, lpData);
	}
	return 0;
}


//------------------------------------------------------------------------
// Changes the current directory in the given dialog box.
// Supported dialog boxes:
// - Standard Windows File/Open and similar: GetOpenFileName, GetSaveFileName
// - MS Office File/Open and similar
// - Standard Windows folder selection: SHBrowseForFolder
//------------------------------------------------------------------------

// Helper class used by setDialogBoxDirectory().
// Should be instantiated every time setDialogBoxDirectory() is called.
class SetDialogBoxDirectory {
public:
	
	// Changes the current directory in the given dialog box.
	//
	// Args:
	//   hwnd: The window handle of the dialog box or one of its controls.
	//   directory: The directory to set in the dialog box.
	//
	// Returns:
	//   True if the dialog box is supported and the current directory has been changed.
	bool doit(HWND hwnd, LPCTSTR directory);
	
private:
	
	// One method per supported dialog box.
	// Sets m_success to true if the directory has been successfully changed.
	//
	// Returns:
	//   True on success, false to continue to the next method.
	bool tryBrowseForFolder();
	bool tryMsOfficeFileOpen();
	bool tryWindowsFileOpen();
	
	void lockWindowUpdate() {
		m_unlock_window_update_needed = true;
		LockWindowUpdate(m_hwnd);
	}
	
	// The dialog box handle, is never a control.
	HWND m_hwnd;
	
	// Directory to set the dialog box to, without any quotes.
	TCHAR m_directory_no_quotes[MAX_PATH];
	
	// Set to true by the set*() methods if LockWindowUpdate(NULL) should be called before returning.
	bool m_unlock_window_update_needed;
};


bool setDialogBoxDirectory(HWND hwnd, LPCTSTR directory) {
	return SetDialogBoxDirectory().doit(hwnd, directory);
}


bool SetDialogBoxDirectory::doit(HWND hwnd, LPCTSTR directory) {
	
	// Retrieves the dialog box handle, supports the case where hwnd is a control.
	while (hwnd && (GetWindowStyle(hwnd) & WS_CHILD)) {
		hwnd = GetParent(hwnd);
	}
	if (!hwnd) {
		return false;
	}
	m_hwnd = hwnd;
	
	lstrcpyn(m_directory_no_quotes, directory, arrayLength(m_directory_no_quotes));
	PathUnquoteSpaces(m_directory_no_quotes);
	
	m_unlock_window_update_needed = false;
	
	// Try all supported dialog boxes until one method returns false.
	bool success = tryBrowseForFolder() || tryMsOfficeFileOpen() || tryWindowsFileOpen();
	
	if (m_unlock_window_update_needed) {
		LockWindowUpdate(NULL);
	}
	
	return success;
}


bool SetDialogBoxDirectory::tryBrowseForFolder() {
	
	// Check for SHBrowseForFolder dialog box:
	// - Must contain a control having a class name containing "SHBrowseForFolder"
	if (!findVisibleChildWindow(m_hwnd, _T("SHBrowseForFolder"), true)) {
		return false;
	}
	
	// Send a BFFM_SETSELECTION message to the dialog box to change the current directory.
	// The messages requires the directory string to be a valid string in the dialog box process:
	// we need to use VirtualAllocEx() and WriteProcessMemory().
	
	// 1) Retrieve a handle to the dialog box process
	DWORD process_id;
	GetWindowThreadProcessId(m_hwnd, &process_id);
	const HANDLE process_handle = OpenProcess(
			PROCESS_VM_OPERATION | PROCESS_VM_WRITE, FALSE, process_id);
	if (!process_handle) {
		return false;
	}
	
	// 2) Allocate a buffer in the process space for the directory path
	const DWORD size = (lstrlen(m_directory_no_quotes) + 1) * sizeof(TCHAR);
	void *const remote_buf = VirtualAllocEx(process_handle, NULL, size, MEM_COMMIT, PAGE_READWRITE);
	bool success = false;
	if (remote_buf) {
		
		// 3) Fill the buffer
		WriteProcessMemory(process_handle, remote_buf, m_directory_no_quotes, size, NULL);
		
		// 4) Send the message using the buffer
		SendMessage(m_hwnd, BFFM_SETSELECTION, TRUE, reinterpret_cast<LPARAM>(remote_buf));
		
		// 5) Free the buffer
		VirtualFreeEx(process_handle, remote_buf, 0, MEM_RELEASE);
		success = true;
	}
	CloseHandle(process_handle);
	return success;
}


bool SetDialogBoxDirectory::tryMsOfficeFileOpen() {
	
	// Check for MS Office File/Open dialog box
	// - Check that the dialog box class contains "bosa_sdm_"
	// - Find the first visible edit box (Edit or RichEdit)
	if (!checkWindowClass(m_hwnd, _T("bosa_sdm_"), true)) {
		return false;
	}
	
	// The file path edit box control ID depends on MS Office version
	const DWORD style = GetWindowStyle(m_hwnd);
	UINT id = 0;
	if (style == 0x96CC0000) {
		id = 0x36;
	} else if (style == 0x94C80000) {
		id = 0x30;
	}
	
	const HWND hwnd_control = GetDlgItem(m_hwnd, id);
	if (!(id && hwnd_control &&
			(checkWindowClass(hwnd_control, _T("Edit"), false) ||
			checkWindowClass(hwnd_control, _T("RichEdit20W"), false)))) {
		return false;
	}
	
	lockWindowUpdate();
	
	// The control is an edit box: save its contents to path_save, then set it to the directory.
	TCHAR path_save[MAX_PATH];
	SendMessage(hwnd_control, WM_GETTEXT, arrayLength(path_save),
			reinterpret_cast<LPARAM>(path_save));
	lockWindowUpdate();
	if (!SendMessage(hwnd_control, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(m_directory_no_quotes))) {
		return false;
	}
	
	// Simulate a press to ENTER to switch to the new directory and wait a bit.
	PostMessage(hwnd_control, WM_KEYDOWN, VK_RETURN, 0);
	PostMessage(hwnd_control, WM_KEYUP, VK_RETURN, 0);
	sleepBackground(100);
	
	// Restore the saved path.
	SendMessage(hwnd_control, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(path_save));
	return true;
}


bool SetDialogBoxDirectory::tryWindowsFileOpen() {
	
	// Check for standard Get(Open|Save)FileName dialog box
	// - Must answer to CDM_GETSPEC message
	if (SendMessage(m_hwnd, CDM_GETSPEC, 0, NULL) <= 0) {
		return false;
	}
	
	// Two possible controls: file path edit box or combo box
	static const UINT control_ids[] = { cmb13, edt1 };
	for (size_t control = 0; control < arrayLength(control_ids); control++) {
		const HWND hwnd_control = GetDlgItem(m_hwnd, control_ids[control]);
		if (!hwnd_control) {
			continue;
		}
		
		// The control is valid: save its contents to path_save, then set it to the directory.
		TCHAR path_save[MAX_PATH];
		SendMessage(hwnd_control, WM_GETTEXT, arrayLength(path_save),
				reinterpret_cast<LPARAM>(path_save));
		lockWindowUpdate();
		if (SendMessage(hwnd_control, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(m_directory_no_quotes))) {
			
			// Simulate a click on the OK button.
			sleepBackground(0);
			SendMessage(m_hwnd, WM_COMMAND, IDOK, 0);
			sleepBackground(0);
			
			// Restore the saved path.
			SendMessage(hwnd_control, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(path_save));
			return true;
		}
	}
	
	return false;
}


//------------------------------------------------------------------------
// Shell API tools
//------------------------------------------------------------------------

// Wrapper for SHGetSpecialFolderPath
bool getSpecialFolderPath(int index, LPTSTR path) {
	return toBool(SHGetSpecialFolderPath(NULL, path, index, TRUE));
}


// Get the target of a file:
// If the file is a shortcut, return its target; handle all known cases:
// - installer shortcuts (MSI API)
// - IUniformResourceLocator shortcuts
// - IShellLink shortcuts
// If the file is a normal file, return its path unchanged
bool getShellLinkTarget(LPCTSTR link_file, LPTSTR target_path) {
	*target_path = _T('\0');
	
	// Resolve Installer shortcuts
	// Use MSI API dynamically, because it is available only under NT-like
	const HMODULE hMSI = LoadLibrary(_T("Msi.dll"));
	if (hMSI) {
		typedef UINT (WINAPI *PFN_MsiGetShortcutTarget)(
			LPCTSTR szShortcutTarget, LPTSTR szProductCode, LPTSTR szFeatureId, LPTSTR szComponentCode);
		const PFN_MsiGetShortcutTarget pfnMsiGetShortcutTarget =
			myGetProcAddress(hMSI, MsiGetShortcutTarget);
		
		TCHAR pszProductCode[39];
		TCHAR pszComponentCode[39];
		if (pfnMsiGetShortcutTarget && ERROR_SUCCESS ==
				pfnMsiGetShortcutTarget(link_file, pszProductCode, NULL, pszComponentCode)) {
			typedef UINT (WINAPI *PFN_MsiGetComponentPath)(
				LPCTSTR szProduct, LPCTSTR szComponent, LPTSTR lpPathBuf, DWORD* pcchBuf);
			const PFN_MsiGetComponentPath pfnMsiGetComponentPath =
				myGetProcAddress(hMSI, MsiGetComponentPath);
			DWORD buf = MAX_PATH;
			if (pfnMsiGetComponentPath) {
				pfnMsiGetComponentPath(pszProductCode, pszComponentCode, target_path, &buf);
			}
		}
		
		FreeLibrary(hMSI);
		
		if (*target_path) {
			return true;
		}
	}
	
	strToW(MAX_PATH, wszLinkFile, link_file);
	
	// Resolve the shortcut using the IUniformResourceLocator and IPersistFile interfaces
	IUniformResourceLocator *purl;
	if (SUCCEEDED(CoCreateInstance(CLSID_InternetShortcut, NULL, CLSCTX_INPROC_SERVER,
			IID_IUniformResourceLocator, (LPVOID*)&purl))) {
		IPersistFile *ppf;
		if (SUCCEEDED(purl->QueryInterface(IID_IPersistFile, (void**)&ppf))) {
			LPTSTR pszURL;
			HRESULT hr1, hr2;
			if (SUCCEEDED(hr1 = ppf->Load(wszLinkFile, STGM_READ)) &&
					SUCCEEDED(hr2 = purl->GetURL(&pszURL))) {
				lstrcpyn(target_path, pszURL, MAX_PATH);
				CoTaskMemFree(pszURL);
			}
			ppf->Release();
		}
		purl->Release();
		
		if (*target_path) {
			return true;
		}
	}
	
	// Resolve the shortcut using the IShellLink and IPersistFile interfaces
	IShellLink *psl;
	if (SUCCEEDED(CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
			IID_IShellLink, (LPVOID*)&psl))) {
		IPersistFile *ppf;
		if (SUCCEEDED(psl->QueryInterface(IID_IPersistFile, (void**)&ppf))) {
			if (SUCCEEDED(ppf->Load(wszLinkFile, STGM_READ)) &&
					SUCCEEDED(psl->Resolve(NULL, MAKELONG(SLR_NO_UI | SLR_NOUPDATE, 1000)))) {
				psl->GetPath(target_path, MAX_PATH, NULL, 0);
			}
			ppf->Release();
		}
		psl->Release();
		
		if (*target_path) {
			return true;
		}
	}
	
	return false;
}


//------------------------------------------------------------------------
// Command line parsing and executing
//------------------------------------------------------------------------

void findFullPath(LPTSTR path, LPTSTR full_path) {
	if (!pathIsSlow(path)) {
		PathUnquoteSpaces(path);
		if (SearchPath(NULL, path, NULL, MAX_PATH, full_path, NULL)) {
			return;
		}
		
		DWORD buf = MAX_PATH;
		if (SUCCEEDED(AssocQueryString(ASSOCF_OPEN_BYEXENAME, ASSOCSTR_EXECUTABLE,
				path, _T("open"), full_path, &buf))) {
			return;
		}
		
		if (32 < reinterpret_cast<UINT_PTR>(FindExecutable(path, NULL, full_path))) {
			return;
		}
	}
	
	lstrcpy(full_path, path);
}


void shellExecuteCmdLine(LPCTSTR command, LPCTSTR directory, int show_mode) {
	// Expand the environment variables before splitting
	TCHAR pszCommandExp[MAX_PATH + bufClipboardString];
	ExpandEnvironmentStrings(command, pszCommandExp, arrayLength(pszCommandExp));
	
	// Split the command line: get the file and the arguments
	TCHAR path[MAX_PATH];
	lstrcpyn(path, pszCommandExp, arrayLength(path));
	PathRemoveArgs(path);
	
	// Get the directory
	TCHAR pszDirectoryBuf[MAX_PATH];
	if (!directory || !*directory) {
		findFullPath(path, pszDirectoryBuf);
		PathRemoveFileSpec(pszDirectoryBuf);
		directory = pszDirectoryBuf;
		PathQuoteSpaces(path);
	}
	
	// Run the command line
	SHELLEXECUTEINFO sei;
	sei.cbSize = sizeof(sei);
	sei.fMask = SEE_MASK_FLAG_DDEWAIT;
	sei.hwnd = e_hwndInvisible;
	sei.lpFile = path;
	sei.lpVerb = NULL;
	sei.lpParameters = PathGetArgs(pszCommandExp);
	sei.lpDirectory = directory;
	sei.nShow = show_mode;
	ShellExecuteEx(&sei);
}


DWORD WINAPI ShellExecuteThread::thread(void* params) {
	ShellExecuteThread* const params_ptr = reinterpret_cast<ShellExecuteThread*>(params);
	shellExecuteCmdLine(params_ptr->m_command, params_ptr->m_directory, params_ptr->m_show_mode);
	delete params_ptr;
	return 0;
}


void skipUntilComma(TCHAR*& chr_ptr, bool unescape) {
	TCHAR* const start = chr_ptr;
	bool escaping = false;
	TCHAR* current = start;
	while (*current) {
		if (escaping) {
			escaping = false;
		} else if (*current == _T(',')) {
			break;
		} else if (unescape && *current == _T('\\')) {
			escaping = true;
		}
		current++;
	}
	if (*current) {
		*current++ = _T('\0');
		while (*current == _T(' ')) {
			current++;
		}
	}
	chr_ptr = current;
	
	if (unescape) {
		const TCHAR* input = start;
		TCHAR* output = start;
		bool escaping = false;
		while (*input) {
			if (*input == _T('\\') && !escaping) {
				escaping = true;
			} else{
				escaping = false;
				*output++ = *input;
			}
			input++;
		}
		*output = _T('\0');
	}
}
